import traceback
import pandas as pd
import numpy as np
from auxiliary_funcs import *
from plot_funcs import *
from collections import namedtuple



def find_period(path, 
tol_norm_diff=10**(-3), 
number_steps=1000,
minimum_number_of_relevant_shifts=2,
minimum_number_of_datapoints_for_correlation_test=300,
minimum_ratio_of_datapoints_for_shift_autocorrelation=0.3,
consider_only_significant_correlation=1,
level_of_significance_for_pearson=0.01,
output_flag=1,
plot_tolerances=1,
reference_time = pd.Timestamp('2017-01-01T12'),):
    '''
    This is the main period detection function. 
    It reads your timeseries from a file given as path and calculates the difference between the original autocorrelation function and several possible shifts in order to find minima which indicate possible periods.
    Then a model is fitted for every suggested period. Afterwards each models performance is evaluated by taking the original time series and subtracting the model. 
    If the model fits the time series well, the leftover should be noise and the autocorrelation function should deteriorate.
    It requires the data path, 
    the tolerance for the norm difference between the unshifted and shifted autocorrelation function for a shift tol_norm_diff,
    the number of times iteratively we increase the tolerance number_steps,
    the minimum number of shifts required for calculation minimum_number_of_relevant_shifts,
    the minimum number of datapoints required for calculation minimum_number_of_datapoints_for_correlation_test,
    the minimum ratio of datapoints for which we calculate the autocorrelation of a shift minimum_ratio_of_datapoints_for_shift_autocorrelation,
    the flag declaring the usage only of correlations matching our criterion consider_only_significant_correlation,
    the minimum significance level for our correlation criterion level_of_significance_for_pearson,
    the output flag setting plotting to on/off output_flag,
    the output flag allowing tolerances to be plotted plot_tolerances,
    a reference time for shift/phase calculation and relevant when fitting the model reference_time.
    The returns are the resulting period res_period, the fitted model res_model if a period was found and a performance criterion res_criteria

    :param path: string
    :param reference_time: pd.Timestamp
    :param tol_norm_diff: positive float
    :param number_steps: positive integer
    :param minimum_number_of_datapoints_for_correlation_test: positive integer
    :param minimum_ratio_of_datapoints_for_shift_autocorrelation: positive float
    :param consider_only_significant_correlation: Boolean
    :param level_of_significance_for_pearson: positive float
    :param output_flag: Boolean
    :param plot_tolerances: Boolean
    :return: positive float, RandomForestRegressor (optional), positive float
    '''

    # Load data
    df_data = pd.read_csv(path, parse_dates=["date"])
    
    # Calculate the autocorrelation function and receive the correlation values r_list, the level of significance list p_list (Step 2 in Algorithm 1 in the paper)
    r_list, p_list, corfunc, lag_list = autocor(df_data["value"], list(range(0,int((df_data["value"].size)-minimum_number_of_datapoints_for_correlation_test))), level_of_significance_for_pearson,consider_only_significant_correlation)

    # Test the datapoints for equidistance
    lag_len = 0
    pw_dist = [y-x for x,y in zip(*[iter(df_data["date"])]*2)]
    min_dist = min(pw_dist)
    max_dist = max(pw_dist)
    if  max_dist == min_dist:
        lag_len = pw_dist[0].total_seconds()/60
        print("Time-equidistant datapoints with a lag size of "+str(lag_len)+" minutes.")
    else:
        print("The datapoints are not time-equidistant!")

    
    if min_dist.total_seconds() <= 0:
        print("Warning: At least two records with the same timestamp!")

    try:
        # Calculate the difference between the unshifted and shifted autocorrelation function for each shift and determine which ones are relevant based on their local minima (Step 3 & 4 in Algorithm 1 in the paper)
        diffs = [shift_diff(i, corfunc) for i in list(range(0,int(np.array(corfunc).size-np.array(corfunc).size*minimum_ratio_of_datapoints_for_shift_autocorrelation)))]
        relevant_diffs, peaks, stop_calculation = get_relevant_diffs(diffs)

        list_relv_pos=[]
        size_list_relv_pos=len(list_relv_pos)

        list_suggested_periods=[]
        list_criterion=[]
        list_norms_data_model=[]
        list_norm_diff_data_model=[]
        list_model_data=[]
        list_tolerances=[]
        list_models=[]
        # stop calculation if no local minima (in addition to at 0) are found (stop_calculation=1)
        if stop_calculation == 0:
            sum_of_shifted_correlation_function = [sum_shifted_function(i, corfunc) for i in list(peaks)]
            df_diffs_lag = pd.DataFrame({'lags': peaks, 'diffs': relevant_diffs, 'sum_of_norms': sum_of_shifted_correlation_function})

            # Step by step extend the set of considered shifts (Step 5 in Algorithm 1 in the paper)
            for tol_for_zero in np.linspace(0,1,number_steps+1):
                # Filter for shifts smaller or equal to our criterion tol_for_zero (Step 5 a) in the paper)
                vec_bool=(df_diffs_lag['diffs']<=tol_for_zero) & (df_diffs_lag['sum_of_norms']>tol_for_zero)
                list_relv_pos=(df_diffs_lag['lags'][vec_bool]).to_list()
                
                if len(list_relv_pos) >= minimum_number_of_relevant_shifts and len(list_relv_pos)>size_list_relv_pos:
                    size_list_relv_pos = len(list_relv_pos)
                    correlationvalues_at_relevant_peaks = np.array(corfunc)[np.array(list_relv_pos)]
                    # If we have no (further) relevant shifts, we can abort (Step 5 b) in Algorithm 1 in the paper)
                    all_relv_pos_with_positive_correlation = sum((correlationvalues_at_relevant_peaks <= 0).astype(int)) <= 0
                    if all_relv_pos_with_positive_correlation==True:
                        list_tolerances.append(tol_for_zero)
                        # Get the time difference between the shifts (Step 5 c) in Algorithm 1 in the paper)...
                        relv_time_diff=((df_data["date"].iloc[list_relv_pos]-df_data["date"].iloc[0]) / pd.Timedelta('1 minutes')).to_list()
                        list_of_periods=np.diff(np.array(relv_time_diff))
                        # ...and calculate their median as suggested period (Step 5 d) in Algorithm 1 in the paper)                    
                        suggested_period = np.median(np.array(list_of_periods)) 
                        suggested_period_in_unit_of_duration_lag=suggested_period 

                        # Fit a model based on the data and the phase inside the period, here calculated using modulo (Step 5 e) in Algorithm 1 in the paper)
                        debug_tmp = (df_data["date"] - reference_time) / pd.Timedelta('1 minutes') % suggested_period_in_unit_of_duration_lag
                        df_data['date_modulo']=(((df_data["date"] - reference_time) / pd.Timedelta('1 minutes')) % suggested_period_in_unit_of_duration_lag).copy()

                        
                
                        
                        model_data,mlp= fit_model(df_data)

                        # Subtract the model data from the original and determine the autocorrelation function as a performance measure (Step 5 f) & g) in Algorithm 1 in the paper)
                        signal_data=df_data["value"].to_numpy().reshape(df_data["value"].size, 1)
                        signal_subtracted_model = signal_data - model_data
                        df_data_difference_signal_model = pd.DataFrame(data=signal_subtracted_model, columns=["value"])
                        # If the output_flag=0 (no graphical output), then calculation time is saved by calculating the autocorrelation function only at the relevant shifts
                        if output_flag==1:
                            r_list_diff, p_list_diff, corfunc_diff, lag_list_diff = autocor(df_data_difference_signal_model["value"], list(range(0,int((df_data_difference_signal_model["value"].size)-minimum_number_of_datapoints_for_correlation_test))),level_of_significance_for_pearson,consider_only_significant_correlation)
                            correlationvalues_signalModel_at_relevant_peaks=np.array(corfunc_diff)[np.array(list_relv_pos)]
                        else:
                            r_list_diff, p_list_diff, cor_func_diff, lag_list_diff = autocor(df_data_difference_signal_model["value"], list_relv_pos, level_of_significance_for_pearson,consider_only_significant_correlation)
                            correlationvalues_signalModel_at_relevant_peaks=np.array(cor_func_diff)
                        reduction_of_correlation = 1 - abs(correlationvalues_signalModel_at_relevant_peaks[1:]).mean() / abs(correlationvalues_at_relevant_peaks[1:]).mean()

                        list_suggested_periods.append(suggested_period)
                        list_criterion.append(reduction_of_correlation)
                        list_model_data.append(model_data)
                        list_models.append(mlp)
                        norm_signal = sum(abs(signal_data))[0] / signal_data.size
                        norm_model = sum(abs(model_data))[0] / model_data.size
                        list_norms_data_model.append(norm_signal + norm_model)
                        norm_diff_between_singal_and_model = sum(abs(signal_subtracted_model))[0] / signal_subtracted_model.size
                        list_norm_diff_data_model.append(norm_diff_between_singal_and_model)
                    else:
                        print('Relevant lag in autocorrelation function with non-positive correlation!')
                        break
            
            
            Results = namedtuple('Results', 'period model criteria')

            # Check if there has any periods been suggested if yes...       
            if len(list_suggested_periods)>0:
                df_periods_criterion=pd.DataFrame({'periods':list_suggested_periods, 'criterion':list_criterion, 'norm_diff':list_norm_diff_data_model ,'sum_norms':list_norms_data_model, 'model_data':list_model_data, 'models':list_models, 'tolerances':list_tolerances})
                period_very_close_fit=df_periods_criterion['periods'][(df_periods_criterion['norm_diff']<=tol_norm_diff) & (df_periods_criterion['sum_norms']>tol_norm_diff)]
                # Test if the fittet model is close to the data (almost perfect fit)
                if period_very_close_fit.size>0:
                    model_very_close_fit = df_periods_criterion['models'][(df_periods_criterion['norm_diff'] <= tol_norm_diff) & (df_periods_criterion['sum_norms'] > tol_norm_diff)]
                    model_data_very_close_fit = df_periods_criterion['model_data'][(df_periods_criterion['norm_diff'] <= tol_norm_diff) & (df_periods_criterion['sum_norms'] > tol_norm_diff)]
                    best_tolerances_close_fit = df_periods_criterion['tolerances'][(df_periods_criterion['norm_diff'] <= tol_norm_diff) & (df_periods_criterion['sum_norms'] > tol_norm_diff)]
                    res_period=period_very_close_fit[0]
                    res_model=model_very_close_fit[0]
                    res_criteria=1.5
                    model_data=model_data_very_close_fit[0]
                    best_tolerance = best_tolerances_close_fit[0]
                    other_tolerances = df_periods_criterion['tolerances'][df_periods_criterion['tolerances'] != best_tolerance].to_list()
                    print('Very small difference between data and model, difference smaller than ' + str(tol_norm_diff))

                    if lag_len:
                        print('The suggested period in minutes is ' + str(res_period) + ', in hours is ' + str(res_period / 60) + ', in days is ' + str(res_period / 60 / 24) + ' and in lags is ' + str(res_period/lag_len))
                    else:
                        print('The suggested period in minutes is ' + str(res_period) + ', in hours is ' + str(res_period / 60) + ' and in days is ' + str(res_period / 60 / 24))
                else:
                    # If the fit is not close but there are suggested periods, then find the period with the best criterion of correlation reduction (Step 6 in Algorithm 1 in the paper)
                    index_min_criterion=df_periods_criterion['criterion'].idxmax()
                    res_period=df_periods_criterion['periods'].iloc[index_min_criterion]
                    res_criteria=df_periods_criterion['criterion'].iloc[index_min_criterion]
                    res_model=df_periods_criterion['models'][index_min_criterion]
                    model_data = df_periods_criterion['model_data'][index_min_criterion]
                    best_tolerance=df_periods_criterion['tolerances'][index_min_criterion]
                    other_tolerances=df_periods_criterion['tolerances'][df_periods_criterion.index != index_min_criterion].to_list()
                    print('Reduction of correlation by model: ' + str(res_criteria) + ' with sigma ' + str(best_tolerance))
                    if lag_len:
                        print('The suggested period in minutes is ' + str(res_period) + ', in hours is ' + str(res_period / 60) + ', in days is ' + str(res_period / 60 / 24) + ' and in lags is ' + str(res_period/lag_len))
                    else:
                        print('The suggested period in minutes is ' + str(res_period) + ', in hours is ' + str(res_period / 60) + ' and in days is ' + str(res_period / 60 / 24))

                if output_flag==1:
                    plot_with_period(df_data, diffs, other_tolerances, best_tolerance, lag_list, r_list, p_list, corfunc, model_data, norm_diff_between_singal_and_model,  plot_tolerances,level_of_significance_for_pearson,consider_only_significant_correlation, minimum_number_of_datapoints_for_correlation_test)
            # If no, plot without period
            else:
                print('List of suggested periods is empty!')
                res_period = -1
                res_criteria = 0
                if output_flag == 1:
                    plot_without_period(df_data, diffs, lag_list, r_list, p_list, corfunc)
                return Results(res_period, None, res_criteria)
        else:
            res_period = -1
            res_criteria = 0
            if output_flag==1:
                plot_without_period(df_data, diffs, lag_list, r_list, p_list, corfunc)
            return Results(res_period, None, res_criteria)
        
        # return the period, the model and the corresponding criterion
        return Results(res_period, res_model, res_criteria)

    except:
        traceback.print_exc()
        res_period = -2
        res_criteria = -2
        return Results(res_period, None, res_criteria)



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from auxiliary_funcs import autocor


def plot_with_period(df_data_aggregated, diffs, other_tolerances, best_tolerance, lag_list, r_list, p_list, corfunc, model_data, norm_diff_between_singal_and_model, plot_tolerances, level_of_significance_for_pearson, consider_only_significant_correlation, minimum_number_of_datapoints_for_correlation_test):
    '''
    This is the regular plotting function in case a period is detected, comparing the real data with the model.
    It plots a comparison of the real data with the one from the model, 
    the norm difference between the unshifted and shifted autocorrelation function (optional: plot possible tolerances and the one with the best performance), 
    the autocorrelation function of the data,
    the difference between data and model (noise) and
    the autocorrelation function of the real data minus the autocorrelation function of the model.
    :param df_data_aggregated: pd.DataFrame
    :param diffs: list of positive floats
    :param other_tolerances: list of positive floats
    :param best_tolerance: positive float
    :param lag_list: list of positive integers
    :param r_list: list of floats
    :param p_list: list of floats
    :param corfunc: list of floats between -1 and  1
    :param model_data: list of 1-D lists of floats
    :param norm_diff_between_singal_and_model: positive float
    :param plot_tolerances: Boolean
    :param level_of_significance_for_pearson: positive float
    :param consider_only_significant_correlation: Boolean
    :param  minimum_number_of_datapoints_for_correlation_test: positive integer
    '''

    signal_data=df_data_aggregated["value"].to_numpy().reshape(df_data_aggregated["value"].size, 1)
    norm_model = sum(abs(model_data))[0] / model_data.size
    norm_signal = sum(abs(signal_data))[0] / signal_data.size
    signal_subtracted_model = signal_data - model_data
    df_data_difference_signal_model = pd.DataFrame(data=signal_subtracted_model, columns=["value"])
    r_list_diff, p_list_diff, corfunc_diff, lag_list_diff = autocor(df_data_difference_signal_model["value"], list(range(0, int((df_data_difference_signal_model["value"].size) - minimum_number_of_datapoints_for_correlation_test))),level_of_significance_for_pearson,consider_only_significant_correlation)
    print('Norm of difference between signal and model: ' + str(norm_diff_between_singal_and_model))
    print('Norm model: ' + str(norm_model))
    print('Norm signal: ' + str(norm_signal))


    plt.plot(df_data_aggregated["value"], label="data")
    plt.plot(model_data, label="model")
    plt.legend()
    plt.title("Data and fitted model")
    plt.xlabel('Time in time points')
    plt.savefig('data_model_trend.png',format='png')
    plt.show()



    plt.plot(list(np.array(diffs)))
    plt.title('Norm of difference between unshifted and shifted autocorrelation function')
    plt.xlabel('Shift in time points')
    plt.ylabel('Function values of d')
    if plot_tolerances == 1:
        plt.hlines(y=other_tolerances, xmin=0, xmax=len(diffs), colors='green')
        plt.hlines(y=[best_tolerance], xmin=0, xmax=len(diffs), colors='red')
    plt.savefig('shift_acf_trend.png',format='png')
    plt.show()

    plt.plot(lag_list, r_list, label="r")
    plt.plot(lag_list, p_list, label="p")
    plt.plot(lag_list, corfunc, label="acf")
    plt.legend()
    plt.title("Autocorrelation function of data")
    plt.xlabel('Shift of the data in time points')
    plt.savefig('acf_trend.png',format='png')
    plt.show()

    plt.plot(df_data_difference_signal_model["value"])
    plt.title("Difference of data and model")
    plt.show()

    plt.plot(lag_list_diff, r_list_diff, label="r")
    plt.plot(lag_list_diff, p_list_diff, label="p")
    plt.plot(lag_list_diff, corfunc_diff, label="acf")
    plt.legend()
    plt.title("Autocorrelation function of data minus model")
    plt.xlabel('Shift of the data minus model in time points')
    plt.savefig('acf_data_model_trend.png',format='png')
    plt.show()

def plot_without_period(df_data_aggregated, diffs, lag_list, r_list, p_list, corfunc):
    '''
    This function plots the results when no period was suggested.
    It shows the time series data,
    the norm difference between unshited and shifted autocorrelation function and 
    the autocorrelation function of the time series.
    :param df_data_aggregated: pd.DataFrame
    :param diffs: list of positive floats
    :param lag_list: list of positive integers
    :param r_list: list of floats
    :param p_list: list of floats
    :param corfunc: list of floats between -1 and  1
    '''
    plt.plot(df_data_aggregated["value"], label="data")
    plt.legend()
    plt.title("time series data")
    plt.show()

    plt.plot(list(np.array(diffs)))
    plt.title('Norm of difference between unshifted and shifted autocorrelation function')
    plt.xlabel('Shift in time points')
    plt.show()

    plt.plot(lag_list, r_list, label="r")
    plt.plot(lag_list, p_list, label="p")
    plt.plot(lag_list, corfunc, label="acfunc")
    plt.xlabel('Time in data points')
    plt.legend()
    plt.title("Autocorrelation function of data")
    plt.show()



import numpy as np
import functools
from scipy.signal import find_peaks
from scipy import stats
from sklearn.ensemble import RandomForestRegressor
from sklearn.svm import SVR



def calculate_autocorrelation(x,i,level_of_significance_for_pearson,consider_only_significant_correlation):
    '''
    This function calculates the autocorrelation function for a vector x and a shift given as index i. 
    The parameters consider_only_significant_correlation and level_of_significance_for_pearson decide if and which results are considered depending on their significance p.
    :param x: list
    :param i: positive integer
    :param level_of_significance_for_pearson: float between 0 and 1
    :param consider_only_significant_correlation: Boolean or interger in {0,1}
    :return: float between -1 and 1, float between 0 and 1, float between -1 and 1, positive integer
    '''
    if i == 0:
        return 1,0,1,0
    else:
        r, p = stats.pearsonr(x[i:], x[:-i])
        if np.isnan(r):
            return 0,0,0,i
        if p > level_of_significance_for_pearson and consider_only_significant_correlation:
            return r,p,0,i
        else:
            return r,p,r,i

def autocor(x,list_of_lags,level_of_significance_for_pearson,consider_only_significant_correlation):
    '''
    This function calculates the autocorrelation function for a time series given as vector x and the shifts given as indices in list_of_lags.
    The parameters consider_only_significant_correlation and level_of_significance_for_pearson decide if and which results are considered depending on their significance p.
    It returns the correlation coefficients as r_list, their significance as p_list, the autocorrelation function as func and the shift indices as lag_list
    :param x: list
    :param list_of_lags: list of positive integers
    :param level_of_significance_for_pearson: float between 0 and 1
    :param consider_only_significant_correlation: Boolean or interger in {0,1}
    :return: list of floats, list of floats, list of floats, list of positive integers
    '''
    list_of_results=list(zip(*list(map(functools.partial(calculate_autocorrelation,x,level_of_significance_for_pearson=level_of_significance_for_pearson,consider_only_significant_correlation=consider_only_significant_correlation),list_of_lags))))
    r_list=list(list_of_results[0])
    p_list=list(list_of_results[1])
    func=list(list_of_results[2])
    lag_list=list(list_of_results[3])
    return r_list, p_list, func, lag_list

def shift_diff(i, corfunc):
    '''
    This function calculates the L^1 norm of the difference of the original autocorrelation function and the shifted version. 
    The values of the autocorrelation function results are given in corfunc and the shift index is given as i.
    :param i: positive integer
    :param corfunc: list of floats between -1 and  1
    :return: positive float
    '''
    if i==0:
        return 0 
    else:
        w1 = np.array(corfunc)[i:]
        w2 = np.array(corfunc)[:-i]
        return float(sum(abs(w1 - w2)) / w1.size)

def fit_model(df_data_aggregated):
    '''
    This function uses the phase of a date in a suggested period as input in order to fit a model. 
    The model will later also test how well besaid period fits the original time series.
    Disclaimer: Here the sklearn.ensemble.RandomForestRegressor() is used, but any model can be used instead.
    :param df_data_aggregated: pd.DataFrame
    :return: list of 1-D lists of floats, RandomForestRegressor object
    '''
    # The routine to fit a model based on the periodic information/phase of a date concerning the period to the original time series to test the hypothesis how well the suggested period fits the original time series
    X = df_data_aggregated["date_modulo"].to_numpy().reshape(df_data_aggregated["date_modulo"].size, 1)
    y = df_data_aggregated["value"].to_numpy().reshape(df_data_aggregated["value"].size)
   
    mlp = RandomForestRegressor()

    mlp.fit(X, y)
    y_model = mlp.predict(X).reshape(X.size, 1)

    return y_model, mlp

def get_relevant_diffs(diffs):
    '''
    This function detects the local minima in the differences diffs between the original autocorrelation function and its shifted versions.
    It returns a list of the peaks, their indices and a Boolean indicating whether further calculation can be aboarted do to a lack of local minima.
    :param diffs: list of positive floats
    :return: list of positive floats, list of positive integers, Boolean
    '''
    stop_calculation=0
    peaks, _ = find_peaks(-np.array(diffs))
    if peaks.size <= 0:
        print('No local minima in the inner domain of the function mapping the L^1 norm of the difference between the autocorrelation function and its shifted versions against the shifts! No period detection possible!')
        stop_calculation = 1
    # For shift=0, two identical functions are subtracted resulting in a local minimum at shift 0 in the function mapping the L^1 norm of the difference between the autocorrelation function and its shifted versions against the shifts.
    # Needs to be inserted separately since find_peaks searches only in the inne of the domain, consequently no boundary values are considered
    peaks = np.insert(peaks, 0, 0)
    peaks = peaks.astype(int)
    return np.array(diffs)[peaks],peaks,stop_calculation

def sum_shifted_function(i, corfunc):
    '''
    This function calculates the sum of the L^1 norm of the correlation function and its shifted version (given as corfunc) for a given shift i.
    This is used to see if their sum is already smaller than our criterion (for further detail see paper Algorithm 1 step 5))
    :param i: positive integer
    :param corfunc: list of floats between -1 and  1
    :return: positive float
    '''
    if i==0:
        return 2*sum(abs(np.array(corfunc))) / np.array(corfunc).size
    else:
        w1 = np.array(corfunc[i:])
        w2 = np.array(corfunc[:-i])
        return sum(abs(w1))/w1.size + sum(abs(w2))/w2.size
